from mpl_toolkits.axes_grid1 import Divider, Size
from mpl_toolkits.axes_grid1.mpl_axes import Axes
from itertools import cycle
import matplotlib as mpl
import matplotlib.pyplot as plt
mpl.rcParams['text.latex.preamble'] = r'\usepackage{amsmath}'
from math import log10, floor
import sortednp as snp
import numpy as np

import warnings

def legend_modulus(ax, loc='lower right', second_legend=False, bbox_to_anchor=0):
    handles = [mpl.lines.Line2D([], [], c='k', label=r'$\mathrm{loss~modulus}$'),
              mpl.lines.Line2D([], [], c='k', mfc='k', label=r'$\mathrm{storage~modulus}$')]


    if second_legend :

        if bbox_to_anchor != 0:
            legend = plt.legend(handles=handles, loc=loc, bbox_to_anchor=bbox_to_anchor)
            ax.add_artist(legend)
        else:
            legend = plt.legend(handles=handles, loc=loc)
            ax.add_artist(legend)
    else:
        if bbox_to_anchor != 0:
            ax.legend(loc=loc, handles=handles, bbox_to_anchor=bbox_to_anchor)
        else :
            ax.legend(loc=loc, handles=handles)

def annotate_crit_shear_rates(ax, gammadotc, etas, tauc_TC, position_txt=0.6):
    gammadot2 = tauc_TC / etas
    ax.axvline(gammadotc, c='green', ls='--', marker='')
    ax.axvline(gammadot2, c='purple', ls='--', marker='')
    ax.annotate(r'$\dot{\gamma}_\mathrm{c}$', xy=(gammadotc, position_txt), c='g',
                xycoords=('data', 'axes fraction'), backgroundcolor='w', ha='center')
    ax.annotate(r'$\sigma_{\mathrm{y}} / \eta_\infty$', xy=(gammadot2, position_txt), c='purple',
                xycoords=('data', 'axes fraction'), backgroundcolor='w', ha='center')

def get_crit_shear_rates(df_fit, sample_id, T):
    df_fith = df_fit[(df_fit.sample_ID == sample_id) & (df_fit.temperature == T)]
    gammadotc = float(df_fith['gamma_dotc'])
    tauc_TC = float(df_fith['tauc_TC'])
    etas = float(df_fith['etas'])
    return gammadotc, tauc_TC, etas

def title_fig(microgel_type, solvant, T, wp=None):
    if wp == None:
        return str(microgel_type) + '_in_' + str(solvant) + '_T' + str(T)
    else :
        return microgel_type + '_' + str(wp).replace('.', 'p') + '_in_' + solvant + '_T' + str(T)

def label_sample(micorgel_type, solvant, T):
    return str(microgel_type) + '~in~' + str(solvant) + '~T' + str(T)

def write_sample_info(ax, microgel_type, T, solvant, w_percent=None):
    if w_percent == None:
        ax.text(0, 1.1, r'$\mathrm{T}~' + str(T) + '^\circ \mathrm{C}~~~~\mathrm{' + microgel_type + '}~\mathrm{in~' + solvant + '} $', va='center', transform=ax.transAxes)
    else:
        ax.text(0, 1.1, r'$\mathrm{T}~' + str(T) + '^\circ \mathrm{C}~~~~\mathrm{' + microgel_type + '}~' + str(w_percent) + '~\%~~~~\mathrm{in~' + solvant + '} $', va='center', transform=ax.transAxes)

def merge_physical_data(x1, x2, y1, y2):
    # Create the variable if they don't exist yet
    try:
        x1
    except NameError:
        x1 = np.array([])
    try:
        y1
    except NameError:
        y1 = np.array([])
    # Change datatype in given arrays in case there are ints or stuff
    x1 = np.asarray(x1, dtype=float)
    x2 = np.asarray(x2, dtype=float)
    y1 = np.asarray(y1, dtype=float)
    y2 = np.asarray(y2, dtype=float)
    # The merge and sort action for x
    merged_x, ind = snp.merge(x1, x2, indices = True)
    # The merge and sort action for y
    ind_1 = ind[0]
    ind_2 = ind[1]
    merged_y = np.zeros(len(merged_x))
    count_1 = 0
    count_2 = 0
    for i in range(0, len(merged_x)):
        if i in ind_1:
            merged_y[i] = y1[count_1]
            count_1 += 1
        elif i in ind_2:
            merged_y[i] = y2[count_2]
            count_2 += 1
    return merged_x, merged_y

def round_physical(x, err_x):
    n = int(floor(log10(abs(err_x))))
    return round(x, -n), round(err_x, -n)

def round_significative(x):
    return round(x, -int(floor(log10(abs(x)))))

def create_plot(two_sided=False, colors = ['#6F4C9B', '#5568B8', '#4D8AC6',
               '#60AB9E', '#77B77D', '#A6BE54',
               '#D1B541', '#E49C39', '#DF4828', '#990D38'], markers = ['o', 'v', '^', 's', 'D', '*'], figsize=(5, 3.4)):
    """
    Crée un environnement de plot

    Parameters
    ----------
    twosided : bool
        allows to change the size of the figure accordingly.
    colors : list of strings
        a default list exists but this allows to change it if u want
    markers : list of strings
        a default list of markers exists, but u can change it if needed
    Returns
    -------
    fig, ax : matplotlib objects to be used as normal

    """
    color = cycle(colors)
    marker = cycle(markers)
    if two_sided :
        fig = plt.figure(figsize=(3.4, 3.4))
    else :
        fig = plt.figure(figsize=figsize)
    # The first & third items are for padding and the second items are for the
    # axes. Sizes are in inches.
    h = [Size.Fixed(1.0), Size.Scaled(1.), Size.Fixed(.2)]
    v = [Size.Fixed(0.7), Size.Scaled(1.), Size.Fixed(.5)]

    divider = Divider(fig, (0.0, 0.0, 1., 1.), h, v, aspect=False)
    # the width and height of the rectangle is ignored.

    ax = Axes(fig, divider.get_position())
    ax.set_axes_locator(divider.new_locator(nx=1, ny=1))

    fig.add_axes(ax)
    return fig, ax, color, marker

def get_fit_amp_sweep(x, y, graph=True):
    warnings.simplefilter("ignore")
    x = x[:, np.newaxis]
    # Fit Ransac
    ransac = sklearn.linear_model.RANSACRegressor(residual_threshold=x.std()/20,
                                                  base_estimator = sklearn.linear_model.LinearRegression(positive=True))
    ransac.fit(x[:], y[:])
    pente = ransac.estimator_.coef_[0]
    while pente != 0:
        ransac.fit(x[:], y[:])
        pente = ransac.estimator_.coef_[0]
    inlier_mask = ransac.inlier_mask_
    outlier_mask = np.logical_not(inlier_mask)
    # Prédiction des deux fits
    line_X = np.linspace(x.min(), x.max(), 100)
    line_X = line_X[:, np.newaxis]
    line_y_ransac = ransac.predict(line_X)



    if graph:
        #print("R²")
        #print(ransac.estimator_.score(line_X, line_y_ransac))
        y0 = ransac.estimator_.intercept_
        slop = ransac.estimator_.coef_[0]

        print("Ordonnée à l'origine", y0)
        print("Pente", slop)
        lw = 1
        plt.scatter(x[inlier_mask], y[inlier_mask], color='yellowgreen', marker='.',
                    label='Inliers')
        plt.scatter(x[outlier_mask], y[outlier_mask], color='gold', marker='.',
                    label='Outliers')
        #plt.plot(line_X, line_y_ransac, color='cornflowerblue', linewidth=lw,
        # label='RANSAC regressor')
        plt.plot(np.linspace(x[0], x[-1], 20),
                y0 + slop * np.linspace(x[0], x[-1], 20),
                lw=1, ls='-', color='teal', label='yo', marker='')
        plt.legend(bbox_to_anchor=(1.1, 1.1))._legend_box.align='left'
        plt.xscale('log')
        plt.yscale('log')
        #plt.xlabel(r'$p~\mathrm{(m)}$')
        #plt.ylabel(r'$F~\mathrm{(N)}$')
        #plt.title('V = ' + str(V))
        return y0, slop


def get_fit_end_recover(x, y, graph=True):
    #warnings.simplefilter("ignore")
    x = x[:, np.newaxis]
    # Fit Ransac
    ransac = sklearn.linear_model.RANSACRegressor(residual_threshold=x.std()/20,
                                                  base_estimator = sklearn.linear_model.LinearRegression(positive=True))
    ransac.fit(x[:], y[:])
    pente = ransac.estimator_.coef_[0]
    inlier_mask = ransac.inlier_mask_
    outlier_mask = np.logical_not(inlier_mask)
    # Prédiction des deux fits
    line_X = np.linspace(x.min(), x.max(), 100)
    line_X = line_X[:, np.newaxis]
    line_y_ransac = ransac.predict(line_X)
    y0 = ransac.estimator_.intercept_
    slop = ransac.estimator_.coef_[0]
    return y0, slop



import numpy as np
from plot_function import create_plot, round_physical, merge_physical_data
from scipy.optimize import curve_fit
import math
import matplotlib as mpl
import pandas as pd
import yaml
from math import log10, floor
import matplotlib.pyplot as plt
mpl.rcParams['text.latex.preamble'] = r'\usepackage{amsmath}'




# define HB model
def HB(x, tauy, K, n):
     return tauy + K*x**n
# define TC model
def TC(x, tauy, gamma_dotc, etas):
    return tauy + tauy*((x / gamma_dotc)**0.5) + etas * x


def fit_TC_HB(df, sample_ID, T, plot=True, replace=False):
    """
    Cette fonction permet de fitter les données d'un dataframe par les deux modèles HB et TC
    Elle prend en entrée :
        df : dataframe
        sample_ID : entier qui correspond à l'échantillon visé
        T : température
    Elle retourne :



    """

    with open('U:\My Documents\Recherche\Données\Samples\samples.yml') as f:
        # use safe_load instead of load
        samples = yaml.safe_load(f)

    w_pourcent = samples[sample_ID]['concentration']
    solvant = samples[sample_ID]['solvant']
    microgel_type = samples[sample_ID]['microgel_type']

    columns = ['sample_ID', 'w_pourcent', 'solvant', 'microgel_type',
                'temperature', 'tauc_HB', 'err_taucHB', 'K', 'err_K', 'n', 'err_n',
                'tauc_TC', 'err_taucTC', 'etas', 'err_etas', 'gamma_dotc',
                'err_gammadotc', 'gammac', 'err_gammac']

    try :
        df_fit_pkl = pd.read_pickle('df_fit.pkl')
    except FileNotFoundError :
        print('new pickle file')
        df_fit_pkl = pd.DataFrame(columns=columns)

    if replace:
        print('data will be replaced')
        df_fit_pkl = pd.DataFrame(columns=columns)

    # Select data of interest in the dataframe
    dfh = df[(df.temperature == T) &
            (df.sample_ID == sample_ID) &
            (df.sweep_direction == 'inverse') &
            (df.experiment_type == 'flow_curve')]

    if len(dfh) == 0:
        print('empty_data_set')
        return 0

    # Merge all stress data

    Stress = np.array([])
    Shear_rate = np.array([])
    for i in dfh.index :
        dfi = dfh.loc[i]
        stress = np.array(dfi['stress'], dtype=float)
        shear_rate = np.array(dfi['shear_rate'], dtype=float)
        Shear_rate, Stress = merge_physical_data(Shear_rate, shear_rate, Stress, stress)
    try:
        # Do the HB fit
        param_bounds_HB=([0,0,0],[np.inf,np.inf,1])
        popt_HB, pcov_HB = curve_fit(HB, Shear_rate, Stress,bounds=param_bounds_HB)
        std_HB = np.sqrt(np.diag(pcov_HB))

        # Extract parameters from HB fit

        # tauc
        tauc_HB, err_taucHB = round_physical(popt_HB[0], std_HB[0])
        #print(tauc_HB, err_taucHB)

        # K
        K, err_K = round_physical(popt_HB[1] , std_HB[1])
        #print(K, err_K)

        # n
        n , err_n = round_physical(popt_HB[2], std_HB[2])
        #print(n, err_n)
    except RuntimeError:
        print('No fit HB possible for sample ', sample_ID)
        plot = False
        return 0


    try:
        # Do the TC fit
        param_bounds_TC=([0,0,0],[np.inf,np.inf,np.inf])

        popt_TC, pcov_TC = curve_fit(TC, Shear_rate, Stress,bounds=param_bounds_TC)
        std_TC = np.sqrt(np.diag(pcov_TC))

        # Extract parameters from TC fit

        tauc_TC, err_taucTC = round_physical(popt_TC[0], std_TC[0])
        #print(tauc_TC, err_taucTC)

        gamma_dotc, err_gammadotc = round_physical(popt_TC[1], std_TC[1])
        #print(gamma_dotc, err_gammadotc)

        test_if_nul, err_etas = round_physical(popt_TC[2], std_TC[2])

        if test_if_nul == 0:
            etas = round(popt_TC[2], int(floor(log10(abs(popt_TC[2])))))
        else:
            etas = test_if_nul
        #print(etas, err_etas)
        if etas == 0:
            print('no high shear regime, no etas')
            plot=False
            return 0
    except RuntimeError:
        print('No fit TC possible for sample ', sample_ID)
        plot = False
        return 0



    # Create pretty labels

    lab_HB = r'$\mathrm{HB} =' + str(tauc_HB) + '\pm' + str(err_taucHB) + '+' + str(K) + '\pm' + str(err_K) + '* \dot{\gamma_\mathrm{c}}^{'  + str(n) + '\pm' + str(err_n) + '}$'
    lab_TC = r'$\mathrm{TC} = ' + str()

    if plot :
        fig, ax, color , marker = create_plot()

        # plot data
        ax.plot(Shear_rate, Stress, c='gray')

        # plot fits
        min_pow = int(math.log10(min(Shear_rate))) -1
        max_pow = int(math.log10(max(Shear_rate)))
        x = np.logspace(-4, 4, base=10, num=30)
        ax.plot(x, HB(x, tauc_HB, K, n), ls='-', c='purple', marker='',
                label=r'$\sigma_\mathrm{y}^\mathrm{HB} + K \dot{\gamma}^n$')
        ax.plot(x, TC(x, tauc_TC, gamma_dotc, etas), ls='-', c='teal', marker='',
                label=r'$\sigma_\mathrm{y}^\mathrm{TC} + \sigma_\mathrm{y}^\mathrm{TC} \left ( \frac{\dot{\gamma}}{\dot{\gamma_\mathrm{c}}} \right ) ^{1/2} + \eta_\infty \dot{\gamma}$')

        # plot boundaries
        ax.axvline(gamma_dotc, ls='--', c='lightgray', marker='')
        ax.text(gamma_dotc, 0.05, r'$\dot{\gamma_\mathrm{c}}$', size='large', c='lightgray', backgroundcolor='w', ha='center')

        ax.axvline(tauc_TC / etas, ls='--', c='lightgray', marker='')
        ax.text(tauc_TC / etas, 0.05, r'$\sigma_\mathrm{y}^{\mathrm{TC}} / \eta_\infty$', size='large', c='lightgray', backgroundcolor='w', ha='center')

        ax.set(xscale='log', yscale='log')
        ax.set(ylim=(0.01, 1e4), xlim=(1e-4, 2e3))
        #ax.legend()
        x_legend = 4e3
        ax.set(xlabel = r'$\dot{\gamma}~\mathrm{(s^{-1})}$', ylabel = r"$\sigma~\mathrm{(Pa)}$")
        ax.text(x_legend, 510, r'$\sigma_\mathrm{y}^\mathrm{HB} = ' + str(tauc_HB) + '\pm' + str(err_taucHB) + '~\mathrm{Pa}$')
        ax.text(x_legend,220, r'$K = ' + str(K) + '\pm' + str(err_K) + '$')
        ax.text(x_legend,80, r'$n = ' + str(n) + '\pm' + str(err_n) + '$')

        ax.text(x_legend, 9, r'$\sigma_\mathrm{y}^\mathrm{TC} = ' + str(tauc_TC) + '\pm' + str(err_taucTC) + '~\mathrm{Pa}$')
        ax.text(x_legend,3, r'$\dot{\gamma_\mathrm{c}} = ' + str(gamma_dotc) + '\pm' + str(err_gammadotc) + '~\mathrm{s^{-1}}$')
        ax.text(x_legend,1, r'$\eta_\infty = ' + str(etas) + '\pm' + str(err_etas) + '~\mathrm{Pa.s}$')

        ax.text(0, 1.05, r'$' + str(w_pourcent) + '~\%$', transform=ax.transAxes)
        #ax.text(0.5, 1.05, r'$\mathrm{' + str(solvant) + '}$', va='center', transform=ax.transAxes)
        #ax.text(0.2, 1.05, r'$\mathrm{T = ~}' + str(T) + '^\circ \mathrm{C}$', transform=ax.transAxes)
        plt.savefig('fits_sample_' + str(sample_ID) + '_T' + str(T) + '.pdf', bbox_inches = 'tight')
        plt.savefig('fits_sample_' + str(sample_ID) + '_T' + str(T) + '.png', bbox_inches = 'tight', dpi=200)

    # Add critical shear rate
    gammac = gamma_dotc *  etas / tauc_TC
    # Error propagation
    err_gammac = err_gammadotc * etas / tauc_TC + err_taucTC * gamma_dotc * etas / (tauc_TC * tauc_TC) + err_etas * gamma_dotc / tauc_TC
    # Remove numbers that are not physical
    gammac, err_gammac = round_physical(gammac , err_gammac)


    dict_fits = {'tauc_HB' : tauc_HB, 'err_taucHB' : err_taucHB,
                              'K': K, 'err_K': err_K, 'n' : n, 'err_n': err_n,
                              'tauc_TC': tauc_TC, 'err_taucTC': err_taucTC,
                              'etas' : etas, 'err_etas': err_etas,
                              'gamma_dotc' : gamma_dotc, 'err_gammadotc': err_gammadotc,
                              'gammac' : gammac, 'err_gammac' : err_gammac}

    print(sample_ID)
    print(dict_fits)


    df_fit_pkl = df_fit_pkl.append(dict(zip(df_fit_pkl.columns,[int(sample_ID), w_pourcent, solvant, microgel_type,
                                                        T, tauc_HB, err_taucHB, K, err_K, n, err_n,
                                                        tauc_TC, err_taucTC, etas, err_etas,
                                                        gamma_dotc, err_gammadotc, gammac, err_gammac])), ignore_index=True)

    df_fit_pkl.to_pickle('df_fit.pkl', compression='infer', protocol=5, storage_options=None)

    return dict_fits



import collections
import os
import io
import xmltodict
import pandas as pd
import numpy as np


def example_emulsion():
    ''' 
    Example flow curve data for oil in water emulsion
    1400 cp mineral oil in 11%Las solution
    temperature 40C 
    '''

    emulsion_75_1400oil_11Las_40C = '''
    999.89475	131.22475
    891.251	121.57225
    794.32825	112.56525
    707.946	104.258
    630.9575	96.58295
    562.3415	89.495925
    501.187	82.973725
    446.68325	76.9523
    398.109	71.44925
    354.815	66.3825
    316.22775	61.738
    281.83825	57.47005
    251.18875	53.562825
    223.872	49.97115
    199.52725	46.678
    177.829	43.6651
    158.49075	40.8936
    141.253	38.34355
    125.8935	35.99815
    112.202	33.8341
    100.000225	31.832825
    89.12585	29.987525
    79.433425	28.279775
    70.7948	26.694725
    63.095775	25.224025
    56.234175	23.86125
    50.118925	22.594175
    44.6684	21.4186
    39.810875	20.322825
    35.4814	19.299625
    31.622875	18.345225
    28.183875	17.461225
    25.11905	16.63715
    22.387175	15.854425
    19.9526	15.1421
    17.782925	14.46635
    15.84895	13.8322
    14.1254	13.239375
    12.5893	12.68375
    11.2202	12.159
    9.99999	11.673425
    8.9125625	11.22175
    7.94331	10.7851
    7.07947	10.38505
    6.3095775	9.998685
    5.62345	9.64747
    5.0118825	9.31164
    4.4668375	8.98708
    3.9810825	8.69169
    3.548155	8.411305
    3.162275	8.146395
    2.81839	7.8957675
    2.5118925	7.65902
    2.2387325	7.4321325
    1.99527	7.221855
    1.778285	7.0237075
    1.5848975	6.8350625
    1.41254	6.6558
    1.258925	6.481415
    1.12203	6.3202925
    0.99999975	6.16896
    0.89125	6.0223275
    0.794334	5.8852825
    0.70794875	5.75524
    0.63096	5.6274075
    0.562345	5.5108375
    0.501192	5.398845
    0.44668775	5.2893075
    0.3981065	5.1846275
    0.354813	5.08736
    0.31623175	4.99462
    0.281839	4.9035125
    0.2511915	4.8152925
    0.2238725	4.7298725
    0.199528	4.6480125
    0.17782975	4.568455
    0.15849025	4.490795
    0.14125325	4.4134175
    0.12589225	4.33758
    0.11220125	4.26175
    0.09999935	4.186445
    0.0891251	4.1115975
    0.0794312	4.0362175
    0.070794875	3.961225
    0.0630968	3.88675
    0.056236125	3.8126625
    0.0501176	3.7392575
    0.044667425	3.666965
    0.039811625	3.5960975
    0.035480175	3.5267725
    0.03162185	3.4587725
    0.028183125	3.393145
    0.025117975	3.3298475
    0.0223869	3.269595
    0.0199527	3.2101825
    0.017782225	3.1534325
    0.015847575	3.0989675
    0.014125525	3.046845
    0.01258835	2.9974
    0.011219275	2.950915
    0.009999548	2.9067975
    '''
    FC = pd.read_csv(io.StringIO(emulsion_75_1400oil_11Las_40C), names=[
                     'Shear rate', 'Stress'], delimiter='\t').astype('float')
    FC['Stress'] = FC['Stress']
    FC.sort_values('Shear rate', ascending=False, inplace=True)

    return FC


def get_data_dict(data_name):
    '''
    parse xml file into dicionary
    '''
    with open(data_name) as xml_file:
        try:
            xml_file.seek(0)
            data_dict = xmltodict.parse(xml_file.read())
        except:
            xml_file.seek(3)
            data_dict = xmltodict.parse(xml_file.read())
    return data_dict


def dicttopanda(datadict):
    ''' Covert dictionary from rheoml to list of pandas table'''

    datasource = datadict['RheoML_Dataset']['ExperimentalData']
    pandalist = []
    counter = 0

    if isinstance(datasource, list) == False:
        datasource = [datasource]
    for experiment in datasource:
        datatable = []

        if 'DMA' in experiment.keys():
            for line in experiment['DMA']:
                linelist = []
                columns = []

                for cellkey in line.keys():
                    linelist.append(float(line[cellkey]['#text']))
                    columns.append(cellkey)
                datatable.append(linelist)
            pandatable = pd.DataFrame(np.array(datatable), columns=columns)

        if 'RVM' in experiment.keys():

            for line in experiment['RVM']:
                linelist = []
                columns = []

                if isinstance(experiment['RVM'], list) == False:
                    line = experiment['RVM']

                for cellkey in line.keys():
                    linelist.append(float(line[cellkey]['#text']))
                    columns.append(cellkey)
                datatable.append(linelist)
            pandatable = pd.DataFrame(np.array(datatable), columns=columns)
        pandalist.append(pandatable)
    return pandalist


class rheology_data(object):
    '''Container for rheology data from trios rheometer software.

    Accept excel file exported with the option multitab.
    The main utility of this class is for test consisting of Multiple
    steps. The rheology_data class accept index and return the nth
    step result as a tuple with (step_name, Pandas Dataframe)

    The rheology_data class can also be added to append the steps of the
    second object to the steps of the first.
    This is usefull to combine consecutive tests when exported as
    multiple files (Rheology advantage software)

    Attributes:
        filename (str): name of data file (xls from trios)

    '''

    def __init__(self, filename, source='trios_multitab_xls'):
        '''
        Assuming Shear stress label 'Stress' And Shear rate label 'Shear rate'
        Args:
            filename (str): name of data file
            source (str): Options 'trios_multitab_xls' (default), 'pandas_export_excel'
        '''
        if source == 'trios_multitab_xls':
            self.filename = str(filename)

            data_file_object = pd.ExcelFile(filename)
            table_name_list = data_file_object.sheet_names

            self.data = collections.OrderedDict()

            for table_name in table_name_list:
                if table_name == 'Details':
                    self.Details = pd.read_excel(data_file_object,
                                                 sheet_name='Details',
                                                 header=None,
                                                 names=['key', 'value']).set_index('key')

                    try:
                        sample_notes = [self.Details.loc['Sample notes'].value]

                        for key, value in self.Details.iloc[self.Details.index.get_loc('Sample notes')+1:self.Details.index.get_loc('Geometry name')].iterrows():
                            sample_notes.append(key)

                        sample_notes = [
                            x for x in sample_notes if str(x) != 'nan']
                        self.sample_notes = sample_notes
                    except:
                        self.sample_notes = ''

                    try:
                        self.instrument_serial = self.Details.loc['Instrument serial number'].value
                    except:
                        self.instrument_serial = ''

                    try:
                        self.geometry_name = self.Details.loc['Geometry name'].value
                    except:
                        self.geometry_name = ''

                    try:
                        self.instrument_type = self.Details.loc['Instrument type'].value
                    except:
                        self.instrument_type = ''

                    try:
                        self.run_date = self.Details.loc['Run date'].value
                    except:
                        self.run_date = None

                else:
                    try:
                        self.data[table_name] = data_file_object.parse(
                            table_name, skiprows=1).drop(0).reset_index().astype('float')
                    except:
                        print('step ' + table_name + ' not loaded')

        elif source == 'pandas_export_excel':
            self.filename = str(filename)

            data_file_object = pd.ExcelFile(filename)
            table_name_list = data_file_object.sheet_names

            self.data = collections.OrderedDict()

            for table_name in table_name_list:
                if table_name == 'Details':
                    self.Details = pd.read_excel(data_file_object,
                                                 sheet_name=table_name)
                else:
                    try:
                        self.data[table_name] = data_file_object.parse(
                            table_name).astype('float')
                    except:
                        print('step ' + table_name + ' not loaded')

        elif source == 'file_like_object':
            self.filename = 'from filelike object'

            data_file_object = pd.ExcelFile(filename)
            table_name_list = data_file_object.sheet_names

            self.data = collections.OrderedDict()

            for table_name in table_name_list:
                if table_name == 'Details':
                    self.Details = pd.read_excel(data_file_object,
                                                 sheet_name='Details',
                                                 header=None,
                                                 names=['key', 'value']).set_index('key')

                    try:
                        sample_notes = [self.Details.loc['Sample notes'].value]

                        for key, value in self.Details.iloc[self.Details.index.get_loc('Sample notes')+1:self.Details.index.get_loc('Geometry name')].iterrows():
                            sample_notes.append(key)

                        sample_notes = [
                            x for x in sample_notes if str(x) != 'nan']
                        self.sample_notes = sample_notes
                    except:
                        self.sample_notes = ''

                    try:
                        self.instrument_serial = self.Details.loc['Instrument serial number'].value
                    except:
                        self.instrument_serial = ''

                    try:
                        self.geometry_name = self.Details.loc['Geometry name'].value
                    except:
                        self.geometry_name = ''

                    try:
                        self.instrument_type = self.Details.loc['Instrument type'].value
                    except:
                        self.instrument_type = ''

                    try:
                        self.run_date = self.Details.loc['Run date'].value
                    except:
                        self.run_date = None

                else:
                    try:
                        self.data[table_name] = data_file_object.parse(
                            table_name, skiprows=1).drop(0).reset_index().astype('float')
                    except:
                        print('step ' + table_name + ' not loaded')

        else:
            raise ValueError('''data not loaded''')

    @property
    def tidy(self):
        for (stepnum, stepdata) in enumerate(self):
            stepdata[1]['Stepnum'] = stepnum
            stepdata[1]['stepname'] = stepdata[0]
            stepdata[1]['filename'] = self.filename
            if stepnum == 0:
                fulldata = stepdata[1]
            else:
                fulldata = pd.concat(
                    [fulldata, stepdata[1]], ignore_index=True, sort=False)
        return fulldata

    def __getitem__(self, i):
        return list(self.data.items())[i]

    def __repr__(self):
        if self.filename is None:
            ret_string = 'No data'
        else:
            ret_string = 'rheology_data('+self.filename+')'
        return ret_string

    def __add__(self, other):
        self.data.update(other.data)
        return self


class data_package(object):
    def __init__(self, data_path, exp_files_dict=None, procedure=None):

        self.procedure = procedure

        if exp_files_dict is None:
            self.source = 'Trios'
            filelist = [item for item in os.listdir(
                data_path) if '.xls' in item]

            data_dict = {}
            for file in filelist:
                data_dict[file] = data_path + file

            self.data_dict = data_dict
            self._len = len(data_dict)

        elif isinstance(exp_files_dict, dict):
            self.source = 'Advantage'
            self.data_dict = exp_files_dict

    @property
    def data_table(self):
        return pd.DataFrame.from_dict(
            {'filename': list(self.data_dict.keys()),
             'filepath': list(self.data_dict.values())})

    def __getitem__(self, index):

        def _concat_rheology_data(list):
            result = list[0]
            for item in list[1:]:
                result += item
            return result

        if isinstance(index, slice):
            return [self[ii] for ii in range(*index.indices(len(self)))]

        elif isinstance(index, int):
            if self.procedure is not None:
                if self.source == 'Trios':
                    return self.procedure(
                        rheology_data(self.data_table.iloc[index]['filepath']))
                if self.source == 'Advantage':
                    return self.procedure(
                        _concat_rheology_data(
                            [rheology_data(item) for item in
                             self.data_table.iloc[index]['filepath']]))

            else:
                if self.source == 'Trios':
                    return rheology_data(self.data_table.iloc[index]['filepath'])
                if self.source == 'Advantage':
                    return self._concat_rheology_data(
                        [rheology_data(item) for item in
                         self.data_table.iloc[index]['filepath']])

    def __len__(self):
        return self._len


if __name__ == "__main__":
    print('ok')
    pass



# -*- coding: utf-8 -*-
"""
Module for rheology data fitting
--------------------------------

Provides functions and fitting procedure to fit rheological data

Currently focusing on flowcurve

Model expressions and LMfit model are provided

Example:

        rheology.rheology_fit.HB -> Model expression (simple function)
        rheology.rheology_fit.HB_model -> lmfit model object

It also provides few convenience functions to rapidly see and plot the result of
the fit.

"""
import lmfit
import pandas as pd
import matplotlib.pyplot as plt
from IPython.display import Math
import numpy as np


def constantstress(x, ystress=0.1):
    """Constant stress model

    Note:

    .. math::
       \sigma=\sigma_y

    Args:

    Returns:
        stress : Shear Stress, [Pa]
    """
    return ystress


# instantiate model class
constantstress_model = lmfit.Model(constantstress, prefix="constantstress_")
""" Lmfit model from equation :meth:`rheofit.models.constantstress`

Note:

"""
constantstress_model.model_expression = Math("\sigma=\sigma_y")


# set parameters for model class
constantstress_model.set_param_hint("ystress", min=0, vary=True)


def Newtonian(x, eta_bg=0.1):
    """Newtonian model

    Note:

    .. math::
       \sigma=\eta_{bg}\cdot\dot\gamma

    Args:

        eta_bg : viscosity [Pa s]

    Returns:
        stress : Shear Stress, [Pa]
    """
    return eta_bg * x


# instantiate model class
Newtonian_model = lmfit.Model(Newtonian, prefix="newtonian_")
""" Lmfit model from equation :meth:`rheofit.models.Bingham`

Note:

Newtonian_model.set_param_hint('eta_bg', min=0, vary=True)


"""
Newtonian_model.model_expression = Math("\sigma=\eta_{bg}\cdot\dot\gamma")


# set parameters for model class
Newtonian_model.set_param_hint("eta_bg", min=0, vary=True)


def Powerlaw(x, n=0.5, K=0.1):
    """Powerlaw model for the stress data

    Note:

    .. math::
       \sigma=K \cdot \dot\gamma^n

    Args:

        K : consistency index [Pa s]
        n : shear thinning index (1 for Newtonian) []

    Returns:
        stress : Shear Stress, [Pa]
    """
    return K * x ** n


# instantiate model class
Powerlaw_model = lmfit.Model(Powerlaw, prefix="PL_")
""" Lmfit model from equation :meth:`rheofit.models.Bingham`

Note:

Powerlaw_model.set_param_hint('K', min=0, vary=True)
Powerlaw_model.set_param_hint('n', min=0, vary=True)


"""
Powerlaw_model.model_expression = Math("\sigma=K\cdot\dot\gamma^n")


# set parameters for model class
Powerlaw_model.set_param_hint("K", min=0, vary=True)
Powerlaw_model.set_param_hint("n", min=0, vary=True)


def Bingham(x, ystress=1.0, eta_bg=0.1):
    """Bingham model

    Note:

    .. math::
       \sigma=\sigma_y+\eta_{bg}\cdot\dot\gamma

    Args:
        ystress: yield stress [Pa]

        eta_bg : Background viscosity [Pa s]

    Returns:
        stress : Shear Stress, [Pa]
    """
    return ystress + eta_bg * x


# instantiate model class
Bingham_model = lmfit.Model(Bingham, prefix="bingham_")
""" Lmfit model from equation :meth:`rheofit.models.Bingham`

Note:

Bingham_model.set_param_hint('ystress', min=0)

Bingham_model.set_param_hint('eta_bg', min=0, vary=True)


"""
Bingham_model.model_expression = Math(
    "\sigma=\sigma_y + \eta_{bg}\cdot\dot\gamma")


# set parameters for model class
Bingham_model.set_param_hint("ystress", min=0)
Bingham_model.set_param_hint("eta_bg", min=0, vary=True)


def TC(x, ystress=1.0, eta_bg=0.1, gammadot_crit=0.1):
    """Three component model

    Note:

    .. math::
       \sigma=\sigma_y+\sigma_y\cdot(\dot\gamma/\dot\gamma_c)^{0.5}+\eta_{bg}\cdot\dot\gamma

    Args:
        ystress: yield stress [Pa]

        eta_bg : Background viscosity [Pa s]

        gammadot_crit : Critical shear rate [1/s]

    Returns:
        stress : Shear Stress, [Pa]
    """
    return ystress + ystress * (x / gammadot_crit) ** 0.5 + eta_bg * x


# instantiate model class
TC_model = lmfit.Model(TC, prefix="TC_")
""" Lmfit model from equation :meth:`rheofit.models.TC`

Note:

TC_model.set_param_hint('ystress', min=0)

TC_model.set_param_hint('eta_bg', min=0, vary=True)

TC_model.set_param_hint('gammadot_crit', min=0)
"""
TC_model.model_expression = Math(
    "\sigma=\sigma_y+\sigma_y\cdot(\dot\gamma/\dot\gamma_c)^{0.5}+\eta_{bg}\cdot\dot\gamma"
)


# set parameters for model class
TC_model.set_param_hint("ystress", min=0)
TC_model.set_param_hint("eta_bg", min=0, vary=True)
TC_model.set_param_hint("gammadot_crit", min=0)


def TCn(x, ystress=1.0, eta_bg=0.1, gammadot_crit=0.1,n=0.5):
    """Three component model

    Note:

    .. math::
       \sigma=\sigma_y+\sigma_y\cdot(\dot\gamma/\dot\gamma_c)^n+\eta_{bg}\cdot\dot\gamma

    Args:
        ystress: yield stress [Pa]

        eta_bg : Background viscosity [Pa s]

        gammadot_crit : Critical shear rate [1/s]
        
        n: Shear thinning exponent

    Returns:
        stress : Shear Stress, [Pa]
    """
    return ystress + ystress * (x / gammadot_crit) ** n + eta_bg * x


# instantiate model class
TCn_model = lmfit.Model(TCn, prefix="TCn_")
""" Lmfit model from equation :meth:`rheofit.models.TC`

Note:

TCn_model.set_param_hint('ystress', min=0)

TCn_model.set_param_hint('eta_bg', min=0, vary=True)

TCn_model.set_param_hint('gammadot_crit', min=0)

TCn_model.set_param_hint('n', min=0, max=1)
"""
TCn_model.model_expression = Math(
    "\sigma=\sigma_y+\sigma_y\cdot(\dot\gamma/\dot\gamma_c)^{0.5}+\eta_{bg}\cdot\dot\gamma"
)


# set parameters for model class
TCn_model.set_param_hint("ystress", min=0)
TCn_model.set_param_hint("eta_bg", min=0, vary=True)
TCn_model.set_param_hint("gammadot_crit", min=0)
TCn_model.set_param_hint("n", min=0, max=1)


def HB(x, ystress=1.0, K=1.0, n=0.5):
    """Hershel-Bulkley Model

    Note:

    .. math::
       \sigma= \sigma_y + K \cdot \dot\gamma^n

    Args:
        ystress: yield stress [Pa]

        K : Consistency index [Pa s^n]

        n : Shear thinning index []

    Returns:
        stress : Shear Stress, [Pa]
    """
    return ystress + K * x ** n


HB_model = lmfit.Model(HB, prefix="HB_")
""" Lmfit model from equation :meth:`rheofit.models.HB`

Note:
HB_model.set_param_hint('ystress', min=0)

HB_model.set_param_hint('K', min=0, vary=True)

HB_model.set_param_hint('n', min=0.0,max=1,vary=True)
"""

HB_model.model_expression = Math("\sigma=\sigma_y+K\cdot\dot\gamma^n")

HB_model.set_param_hint("ystress", min=0)
HB_model.set_param_hint("K", min=0, vary=True)
HB_model.set_param_hint("n", min=0.0, max=1, vary=True)


def casson(x, ystress=1.0, eta_bg=0.1):
    """Casson Model

    Note:

    .. math::
       \sigma^{0.5}= \sigma_y^{0.5} + \eta_{bg}^{0.5}

    Args:
        ystress: yield stress [Pa]

        eta_bg : Background viscosity [Pa s]

    Returns:
        stress : Shear Stress, [Pa]
    """
    return (ystress ** 0.5 + (eta_bg * x) ** 0.5) ** 2


casson_model = lmfit.Model(casson, prefix="casson_")
""" Lmfit model from equation :meth:`rheofit.models.casson`

Note:
casson_model.set_param_hint('ystress', min=0)

casson_model.set_param_hint('eta_bg', min=0, vary=True)

"""
casson_model.model_expression = Math(
    "\sigma^{0.5}=\sigma_y^{0.5}+\eta_{bg}^{0.5}")

casson_model.set_param_hint("ystress", min=0)
casson_model.set_param_hint("eta_bg", min=0, vary=True)


def carreau(x, eta_0=1.0, gammadot_crit=1.0, n=0.5, prefix="carreau"):
    """carreau Model

    Note:

    .. math::
       \sigma=\dot\gamma \cdot \eta_0 \cdot (1+(\dot\gamma/\dot\gamma_c)^2)^{(n-1)/2}

    Args:
        eta_0: low shear viscosity [Pa s]

        gammadot_crit: critical shear rate [1/s]

        n : shear thinning exponent

    Returns:
        stress : Shear Stress, [Pa]
    """
    return x * eta_0 * (1 + (x / gammadot_crit) ** 2) ** ((n - 1) / 2)


carreau_model = lmfit.Model(carreau, prefix="carreau_")
""" Lmfit model from equation :meth:`rheofit.models.carreau`

Note:
carreau_model.set_param_hint('eta_0', min=0)

carreau_model.set_param_hint('gammadot_crit_carreau', min=0, vary=True)

carreau_model.set_param_hint('n',min=0, max=1)
"""
carreau_model.model_expression = Math(
    "\sigma=\dot\gamma \cdot \eta_0 \cdot (1+(\dot\gamma/\dot\gamma_{c_carreau})^2)^{(n-1)/2}"
)

carreau_model.set_param_hint("eta_0", min=0)
carreau_model.set_param_hint("gammadot_crit", min=0, vary=True)
carreau_model.set_param_hint("n", min=0, max=1)


def cross(x, eta_inf=0.001, eta_0=1.0, n=0.5, gammadot_crit=1.0):
    """cross Model

    Note:

    .. math::
       \sigma= \dot\gamma \eta_{inf} + \dot\gamma (\eta_0 - \eta_{inf})/(1 + (\dot\gamma/\dot\gamma_c)^n)

    Args:
        ystress: yield stress [Pa]

        K : Consistency index [Pa s^n]

        n : Shear thinning index []

    Returns:
        stress : Shear Stress, [Pa]
    """
    return x * eta_inf + x * (eta_0 - eta_inf) / (1 + (x / gammadot_crit) ** n)


cross_model = lmfit.Model(cross, prefix="cross_")
""" Lmfit model from equation :meth:`rheofit.models.HB`

Note:
cross_model.set_param_hint('eta_0', min=0)

cross_model.set_param_hint('eta_inf', min=0, vary=True)

cross_model.set_param_hint('n', min=0.0,max=1,vary=True)

cross_model.set_param_hint('gammadot_crit', min=0.0 ,vary=True)

"""

cross_model.model_expression = Math(
    "\sigma= \dot\gamma \eta_{inf} + \dot\gamma (\eta_0 - \eta_{inf})/(1 + (\dot\gamma/\dot\gamma_c)^n)"
)

cross_model.set_param_hint("eta_0", min=0)
cross_model.set_param_hint("eta_inf", min=0, vary=True)
cross_model.set_param_hint("n", min=0.0, max=1, vary=True)
cross_model.set_param_hint("gammadot_crit", min=0.0, vary=True)


def show_parameter_table(result):
    """ Convenience function to convert lmfit fit parameter result information
        a panda dataframe for easy display and concatenation.

        Args:

        result: Lmfit.fitresult object

        Returns:

        Pandas datafrme with estimated parameters and some quality of fit metrics
    """
    table_list = (
        pd.DataFrame.from_dict(result.params.valuesdict(), orient="index")
        .rename(columns={0: result.model.name})
        .transpose()
    )
    # table_list['aic']=result.aic
    table_list["bic"] = result.bic
    # table_list['chisqr']=result.chisqr
    table_list["redchi"] = result.redchi

    return table_list


def fit_FC(model, data):
    """ Convenience function to fit a flow curve
        Args:

        model: rheology model (e.g. HB_model)

        data: pandas DataFrame with column 'Shear rate' and 'Stress'

        Returns:

        lmfit.fitresult
    """
    return model.fit(data["Stress"], x=data["Shear rate"], weights=1 / data["Stress"])


def plot_fit_fc(result, show_table=True):
    """ Convenience function to plot data and fit of a flow curve
        Args:

        result: lmfit.fitresult

        show_table: default=True (show fit parameter table)

        Returns:

        None and display plot on current ax
    """

    kwarg = {"yscale": "log", "xscale": "log"}
    result.plot_fit(ax_kws=kwarg, yerr=False)
    plt.xlabel("Shear rate [1/s]")
    plt.ylabel("Stress [Pa]")
    if show_table:
        display(show_parameter_table(result))



import pandas as pd
import yaml
import numpy as np
from plot_function import *
from scipy.optimize import curve_fit
import glob

# define TC model
def TC(x, tauy, gamma_dotc, etas):
    return tauy + tauy*((x / gamma_dotc)**0.5) + etas * x

def fit_TC(df):
    df['tauc_TC'] = np.nan ; df['err_taucTC'] = np.nan
    df['gamma_dotc'] = np.nan ; df['err_gammadotc'] = np.nan
    df['etas'] = np.nan ; df['err_etas'] = np.nan
    fig, ax, color, m = create_plot()
    ax.set(xscale='log', yscale='log',
           xlabel=r'$\dot{\gamma}~\mathrm{(1/s)}$',
           ylabel=r'$\tau~\mathrm{(Pa)}$')
    for i in df.index:
        c=next(color)
        dfi = df.loc[i]
        stress = np.array(dfi['stress'], dtype=float)
        #phi=dfi.phi
        shear_rate = np.array(dfi['shear_rate'], dtype=float)


        ax.plot(shear_rate, stress, c=c)
        try:
            # Do the TC fit
            param_bounds_TC=([0,0,0],[np.inf,np.inf,np.inf])

            popt_TC, pcov_TC = curve_fit(TC, shear_rate, stress,bounds=param_bounds_TC)
            std_TC = np.sqrt(np.diag(pcov_TC))

            # Extract parameters from TC fit

            tauc_TC, err_taucTC = round_physical(popt_TC[0], std_TC[0])
            df.at[i, 'tauc_TC'] = tauc_TC
            df.at[i, 'err_taucTC'] = err_taucTC


            gamma_dotc, err_gammadotc = round_physical(popt_TC[1], std_TC[1])
            df.at[i, 'gamma_dotc'] = gamma_dotc
            df.at[i, 'err_gammadotc'] = err_gammadotc

            etas, err_etas = round_physical(popt_TC[2], std_TC[2])
            df.at[i, 'etas'] = etas
            df.at[i, 'err_etas'] = err_etas

            # plot the fits to check if they make sense
            ax.plot(shear_rate, TC(shear_rate, tauc_TC, gamma_dotc, etas ), ls='--', c=c, marker='')
        except RuntimeError:
            print('No fit TC possible')

csv_files = glob.glob('*.csv')


papers_info = yaml.safe_load(open(glob.glob('data*.yml')[0]))
for file in csv_files:
    name = file.replace('.csv', '')

    df_raw = pd.read_csv( name + '.csv')

    df = pd.DataFrame(columns=['ID', 'solvent_viscosity', 'phi', 'shear_rate', 'stress'])
    ID = name
    solvent_viscosity = papers_info[name]['solvent_viscosity']
    for i, column_name in enumerate(df_raw.columns):
        if i%2 == 0:
            phi = float(column_name)
            if papers_info[name]['unit_x'] == 's-1':
                shear_rate = np.array(df_raw[column_name], dtype=float)
            elif papers_info[name]['unit_x'] == '0.158s-1':
                shear_rate = np.array(df_raw[column_name], dtype=float) / 0.158
            shear_rate = shear_rate[~np.isnan(shear_rate)]
        else:
            if papers_info[name]['unit_y'] == 'Pa':
                stress =  np.array(df_raw[column_name])
            elif papers_info[name]['unit_y'] == 'dynes/cm2':
                stress = 0.1* np.array(df_raw[column_name])
            elif papers_info[name]['unit_y'] == '0.0825Pa':
                stress =  np.array(df_raw[column_name]) * 0.0825
            stress = stress[~np.isnan(stress)]



            df = df.append(dict(zip(df.columns,
                                    [ID, solvent_viscosity, phi, shear_rate, stress])),
                                    ignore_index=True)

    fit_TC(df)
    df.to_pickle('df_' + name + '.pkl')



# -*- coding: utf-8 -*-

# Learn more: https://gitlab.pg.com/caggioni.m/rheology_toothpaste

from setuptools import setup, find_packages

readme='Python labrary for rheology data analysis'
licence='MIT'

setup(
    name='rheofit',
    version='0.1.0',
    description='Python library for rheology data analysis',
    long_description=readme,
    author='Marco Caggioni',
    author_email='marco.caggioni@gmail.com',
    url='https://github.com/rheopy/rheofit',
    license=licence,
    install_requires=['lmfit','xmltodict','emcee','corner','pybroom'],
    packages=['rheofit']
)



import ipywidgets as widgets
import numpy as np
import matplotlib.pyplot as plt
import lmfit

def make_par_widget(model,data=None):
    ''' '''
    if data is not None:
        res_fit=model.fit(data['Stress'],x=data['Shear rate'],weights=1/(0.05*data['Stress']))
        params=res_fit.params
    else:
        params=model.make_params()
    
    
    
    model_name_widget=widgets.Text(description='Model name',value=model.name,
                                   style={'description_width': 'initial'},
                                   layout=widgets.Layout(width='80%'))
    
    par_list_wid=[]
    for name,param in params.items():    
        par_list_wid.append(widgets.HBox([widgets.Text(value=name),
                                          widgets.Text(description='min/val/max',value=str(param.min)),
                                          widgets.Text(value=str(param.value)),
                                          widgets.Text(value=str(param.max)),
                                          widgets.Checkbox(description='fix',value=not(param.vary)),
                                         ]))
                      
    box_layout = widgets.Layout(display='flex',
                                flex_flow='column',
                                align_items='stretch',
                                border='solid')
    
    par_widgets=widgets.VBox([model_name_widget]+par_list_wid,layout=box_layout)
    display(par_widgets)
    return par_widgets

def update_par_widget(par_wid,fit_res):
    params=fit_res.params
    
    for param, wid in zip(params.values(),par_wid.children[1:]):
        wid.children[2].value=str(param.value)

def make_par_from_widget(par_widget):
    params = lmfit.Parameters()
    
    for par_info in par_widget.children[1:]:
        
        params.add(par_info.children[0].value,
                      min=float(par_info.children[1].value),
                      value=float(par_info.children[2].value),
                      max=float(par_info.children[3].value),
                      vary=not(par_info.children[4].value),
                      )

    return params


def plot_fit_res(fit_res, show_par_values=False,exp_err=0.05):
    
    fig, (ax1,ax3) = plt.subplots(2,1, sharex=True)
    ax2 = ax1.twinx()

    ax1.plot(fit_res.userkws['x'],fit_res.data,'o', color='red', mfc='none', label='data')
    ax1.plot(fit_res.userkws['x'],fit_res.eval(x=fit_res.userkws['x']),'-', color='black', label='best fit')

    ax2.plot(fit_res.userkws['x'],fit_res.data/fit_res.userkws['x'],'o', color='blue', mfc='none')
    ax2.plot(fit_res.userkws['x'],fit_res.eval(x=fit_res.userkws['x'])/fit_res.userkws['x'],'-', color='black')
    
    ax3.plot(fit_res.userkws['x'], (fit_res.data-fit_res.eval(x=fit_res.userkws['x']))/fit_res.data,'o', color='blue', mfc='none')
    ax3.fill_between(fit_res.userkws['x'], -exp_err,exp_err,
                     color='blue',alpha=0.2,label='estimated exp error')

    
    ax1.set_yscale('log')
    ax1.set_xscale('log')
    ax2.set_yscale('log')

    ax1.set_ylabel('$\sigma [Pa]$')
    ax3.set_xlabel('$\dot\gamma [1/s]$')
    ax2.set_ylabel('$\eta [Pa s]$')
    ax3.set_ylabel('relative residuals')
    
    ax3.set_ylim(-0.2,0.2)
    
    ax3.legend()
    
    if show_par_values:
        mod_par_text=''
        for item in fit_res.params:
            mod_par_text+=(f'{item} : {fit_res.params[item].value:.2E} \n')

        mod_par_text+=f'Red chi square: {fit_res.redchi:.2E} \n'

        plt.text(-1, 0.95, mod_par_text, transform=plt.gca().transAxes, fontsize=14,verticalalignment='top')
        fig.suptitle(fit_res.model)
    
    return fig

def plot_confidence(res_fit,expand=1):
    dely = res_fit.eval_uncertainty(x=res_fit.userkws['x'],sigma=3)*expand

    plt.plot(res_fit.userkws['x'], res_fit.data,'o',color='black',label='Data',markersize=5)
    plt.plot(res_fit.userkws['x'], res_fit.best_fit,label='Best fit TC model',color='red')
    plt.fill_between(res_fit.userkws['x'], res_fit.best_fit-dely,res_fit.best_fit+dely,
                     color='blue',alpha=0.2,label='0.9973 Confidence interval')
    plt.yscale('log')
    plt.xscale('log')
    plt.ylabel('$\sigma [Pa]$')
    plt.xlabel('$\dot\gamma [1/s]$')
